!  Last change:  ZK    7 Nov 2018    2:04 pm
 

      PROGRAM slosh
!  THIS PROGRAM  COMPUTES HYDROSTATIC AND NON-HYDROSTATIC PRESSURE COMPONENTS
!  IN TSUNAMI PROPAGATION. DEPTH IS GIVEN IN THE FILES ale2min.dat AND  ale2minf.dat
!  THE LATTER IS AVERAGED DEPTH AND IS USED TO CALCULATE THE NON-HYDROSTATIC PRESSURE (Q)
!  THE TSUNAMI SOURCE IS GIVEN IN THE FILE source_2min.dat FOR THE 2011 JAPAN TSUNAMI.
!  THE HYDROSTATIC PART IS SOLVED BY THE SUBROUTINES VELOCITYX,  VELOCITYY, CONTX AND CONTY.
!  THE NON-HYDROSTATIC PART INVOKES THE SUBROUTINES VERT_BOTTOM TO OBTAIN VERTICAL VELOCITY AT THE BOTTOM
!  AND NON_HYDRO_CGSTAB WHICH ORGANIZES COMPUTATIONS, FIRST BY ESTABLISHING COEFFICIENTS IN AN ELLIPTICAL
!  EQUATION AND THEN BY CALLING SUBROUTINE CGSTAB2D TO OBTAIN NEW VALUES OF THE NON-HYDROSTATIC PRESSURE - !  (Q),
! AND FINALLY COMPUTING THE VERTICAL VELOCITY AT THE SURFACE AND THE HORIZONTAL COMPONENTS OF
!  VELOCITY  CREATED BY PRESSURE Q.


!  HYDROSTATIC PART OF THE PROGRAM IS WRITTEN BY Z. K.

!  Non-hydrostatic Pressure solver in Spherical coordinates
!  originated by  Z.K.,  IMS, University of Alaska Fairbanks
!  written  by    Y.Yamazaki 2007  ORE, University of Hawaii Manoa
! Program was tested  by W.K. 2018.
      INCLUDE "block2d_2min"


!  Note -- all units are in CGS (Set density = 1)

!          V  |    V |  V  |
!   K+1
!         --  +  --  + --  +      SPACE LOCATION OF SEA LEVEL AND VELOCITY
!         U   Z  U   Z  U  Z
!          J-1     J    J+1       GRID ENUMERATION
!          V  |    V |  V  |
!   K
!         --  +  --  + --  +      SPACE LOCATION OF SEA LEVEL AND VELOCITY
!         U   Z  U   Z  U  Z
!          J-1     J    J+1       GRID ENUMERATION
  
      INTEGER,PARAMETER :: IS=34
      REAL, DIMENSION(1:180, 1:240) :: ETS1
      integer, dimension(IS) :: NUMS, JMS, KMS
      INTEGER :: i, j1, k1, UNITS
      REAL, DIMENSION(IS) :: LONGS, LATS, ZSU
      CHARACTER*14, DIMENSION(IS) :: SNAME
      CHARACTER*16 :: NAMEs

!  Initializations
      ZSU=0.
      DHU=0.; DHV=0.
      M=0; IC=0; KK=0
      CELL=0; CELLX=0
      VO = 0.0; VN=0.; SLNN=0.

      UO=0.; UN=0.; SLO=0.; SLN=0.; D=0.; H=0.
      ZINP=0.
      SLMA=0.

      ETO=0; ETN=0
      JM=0;KM=0

      ZMAX=-1.

      WBO=0.
      WBN=0.
      WSN=0.
      WSO=0.
      SIZ=0.; SIV=0.; CSZ=0.; CSV=0.; CSVZ=0.; CSVZ1=0.
      Q=0.


      open(unit=14,file='long1u.dat',status='unknown')
      open(unit=15,file='long1.dat',status='unknown')
     ! open(unit=16,file='surface.dat',status='unknown')
      open(unit=16,file='surface.dat',form='unformatted',status='unknown')

     ! open(unit=17,file='u.dat',status='unknown')
      open(unit=17,file='u.dat',form='unformatted',status='unknown')

     ! open(unit=18,file='v.dat',status='unknown')
      open(unit=18,file='v.dat',form='unformatted',status='unknown')

     ! open(unit=19,file='uen.dat',status='unknown')
      open(unit=19,file='uen.dat',form='unformatted',status='unknown')

     ! open(unit=20,file='ven.dat',status='unknown')
      open(unit=20,file='ven.dat',form='unformatted',status='unknown')
      open(unit=21,file='ampmax.dat',form='unformatted',status='unknown')
      open(unit=25,file='ale2min.dat',form='unformatted',status='unknown')
      open(unit=26,file='ale2minf.dat',form='unformatted',status='unknown')
      open(unit=40,file='en2min_reg129_west.dat',status='unknown')
      open(unit=41,file='en2min_reg39.5_north.dat',status='unknown')
      open(unit=42,file='en2min_reg44.5_north.dat',status='unknown')
      open(unit=43,file='uenq.dat',form='unformatted',status='unknown')
      open(unit=44,file='venq.dat',form='unformatted',status='unknown')

!FILES FOR RECORDING TIME VARIABILITY
!IS=28
      open(unit=30,file='station_aleutian2min.dat',status='unknown')

      DO I=1,IS
        READ(30,*)NUMS(I), Jms(I), KMS(I), LONGS(I), LATS(I), SNAME(I)
      END DO
      PRINT*,SNAME(1),JMS(1),SNAME(32),KMS(32)
 
      DO I=1, IS
        UNITS=100+i
        WRITE(names,'(A7, I2.2, A4)')'station',i,'.dat'
        open(UNIT=uniTS,file=names,status='unknown')
      END DO

 ! INITIAL TIME STEPPING


 ! ESTABLISH DEPTH AND SEA LEVEL

      CALL DEPTH4_RIGHT()

  !    GO TO 203
      DO I=1, IS
        UNITS=100+i
        WRITE(UNITS,*)NUMS(I), Jms(I), KMS(I), LONGS(I), LATS(I), SNAME(I), H(JMS(I), KMS(I))
      END DO

  !    PRINT*,'H(1250,500)=',H(2500,1000)
  !    GO TO 203

      CALL TRIGON()
      open(unit=27, file='source_2min.dat',status='unknown')
      ETS = 0.
      DO k1=1, 240
        READ(27,*)(ETS1(J1, K1),J1=1,180)
      END DO

      DO k1=1, 240
        K=K1+490
        DO J1=1, 180
          J=580+J1
          ETS(J,K)=ETS1(J1, K1)
        END DO
      END DO

      DO K=1,KE
        DO J=1,JE
          IF (H(J,K) .LE. 0.) ETS(J, K)=0.            ! Islands no down
          SLO(J, K)=ETS(J, K)
        END DO
      END DO
      PRINT *, 'SLO(650,600)=',SLO(650,600)

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

! Starting main time loop

      DO M=1, ME    !TIME LOOP

! RADIATING BOUNDARY CONDITIONS
        CALL NORTH_RADIATING()
        CALL SOUTH_RADIATING()
        CALL RIGHT_RADIATING()
        CALL LEFT_RADIATING()

       ! CALL RIGHT_FORCING()

!--ESTABLISH WET-DRY-------------
        D = H+SLO-ETO
        DO K=KS, KE 
          DO j=js, je
            IF (D(J, K) .LT. EPSILON) then
              SLO(J, K)=-H(J, K)+ETO(J, K)
              CELL(J, K)=0
            ELSE IF (D(J, K) .GE. EPSILON) then
              CELL(J, K)=1
            END IF
          END DO
        END DO

! THIS IS ACTUALLY RUNUP  ALONG X DIRECTION
        CELLX= 0    !TEMPORARY VALUE IF NEW CELL IS GENERATED IN THE RUNUP PROCESS
                    !ALONG X DIRECTION IT IS NOT USED FOR THE RUNUP ALONG Y. ONLY OLD CELLS
                    ! ARE USED IN THE PROCESS.

        DO K=1,KE
          DO j = JS+1, JE-1
            IF (CELL(J, K) .EQ. 1 .AND. CELL(J+1, K) .EQ. 0) then
              if ((SLO(J, K) + H(J+1, K)-ETO(J+1, K)) .GT. epsilon) then
                D(J+1, K)= SLO(J, K) + H(J+1, K)-ETO(J+1, K)
                CELL(J+1, K) = 1
                CELLX(J+1, K) = 1
 !               SLO(J+1,K) = SLO(J,K)          !
                if (CELL(J-1, K) .EQ. 1) then
                  UO(J+1, K)=UO(J, K)
                end if
                if (CELL(J-1, K) .EQ. 0) then   ! CELL (J,K) IS SURROUNDED BY
                  UO(J+1, K)=SQRT(G*D(J+1, K))  ! DRY CELLS
                end if
              end if
            END IF

            IF (CELL(J-1, K) .EQ. 0 .AND. CELL(J, K) .EQ. 1) then
              if ((SLO(J, K) + H(J-1, K)-ETO(J-1, K)) .GT. EPSILON) then
                D(J-1, K)= SLO(J, K) + H(J-1, K)-ETO(J-1, K)
                CELL(J-1, K) = 1
                CELLX(J-1, K) = 1
 !               SLO(J-1, K) = SLO(J, K)        !
                if (CELL(J+1, K) .EQ. 1)then
                  UO(J, K)=UO(J+1, K)
                end if
                if (CELL(J+1, K) .EQ. 0) then   !CELL(J,K) SURROUNDED BY DRY CELLS
                  UO(J, K)=-SQRT(G*D(J-1, K))
                end if
              end if
            END IF
          END DO
        END DO

! THIS IS ACTUALLY RUNUP ALONG Y DIRECTION
        DO K = KS+1, KE-1        
          DO J=1,JE
            IF (Cell(J, K) .EQ. 1 .AND. CELL(J, K+1) .EQ. 0) then
              if ((SLO(J, K) + h(J, K+1)-ETO(J, K+1)) .GT. EPSILON .AND. CELLX(J, K) .EQ. 0) then
                D(J, K+1)= SLO(J, K) + h(J, K+1)-ETO(J, K+1)
                Cell(J, K+1) = 1
!                SLO(J, K+1) = SLO(J, K)
                if (CELL(J, K-1) .EQ. 1) then
                  VO(J, K)=VO(J, K-1)
                end if
                if (CELL(J, K-1) .EQ. 0) then
                  VO(J, K)=SQRT(G*D(J, K+1))
                end if
              end if
            END IF

            IF (CELL(J, K-1) .EQ. 0 .AND. CELL(J, K) .EQ. 1) then
              if ((SLO(J, K) + h(J, K-1)-ETO(J, K-1)) .GT. EPSILON .AND. CELLX(J, K) .EQ. 0) then
                D(J, K-1) = SLO(J, K)+ h(J, K-1)-ETO(J, K-1)
                Cell(J, K-1) = 1
 !               SLO(J, K-1) = SLO(J, K)
                if (CELL(J, K+1) .EQ. 1) then
                  VO(J,K-1)=VO(J,K)
                end if
                if (CELL(J, K+1) .EQ. 0) then
                  VO(J, K-1)=-SQRT(G*D(J, K-1))
                end if
              end if
            END IF
          END DO
        END DO

! SETTING VELOCITY ALONG X
        DO K=KS+1,KE-1
          DO J=JS+1,JE-1
            IF (CELL(J, K) .EQ. 0 .AND. Cell(j-1, K) .EQ. 0) then
              UO(J, K)=0.
            END IF
          END DO

          DO j = JS+1, JE - 1                                             ! CHANGE 1  TO 2 FOR RADIATION
            IF (Cell(j, K) .EQ. 1 .AND. CELL(J+1, K) .EQ. 0) then
              if ((SLO(j, K) + h(j+1, K)-ETO(J+1, K)) .LE. EPSILON) then  ! FOR THE BOUNDARY CONDITION U=0
                UO(J+1, K)=0.                                             ! AT THE RIGHT WALL
              end if
            END IF
            IF (CELL(J-1, K) .EQ. 0 .AND. CELL(J, K) .EQ. 1) then
              if ((SLO(j, K) + h(j-1, K)-ETO(J-1, K)) .LE. EPSILON) then  ! FOR THE BOUNDARY CONDITION U=0
                UO(J,K)=0.                                                ! AT THE LEFT WALL
              end if
            END IF
          END DO
        END DO

!  SETTING VELOCITY ALONG Y
        DO J=JS,JE-1       
          DO K=KS+1,KE-1        
            IF (CELL(J, K) .EQ. 0 .AND. CELL(J, K+1) .EQ. 0)then
              VO(J, K)=0.
            END IF
          END DO

          DO K = KS+1, KE-1
            IF (Cell(J, K) .EQ. 1 .AND. CELL(J, K+1) .EQ. 0) then
              if ((SLO(J, K) + h(J, K+1)-ETO(J, K+1)) .LE. EPSILON) then   ! FOR THE BOUNDARY CONDITION V=0
                VO(J, K)=0.                                                 ! AT THE RIGHT WALL
              end if
            END IF
            IF (CELL(J, K-1) .EQ. 0 .AND. CELL(J,K) .EQ. 1) then
              if ((SLO(J, K) + h(J, K-1)-ETO(J, K-1)) .LE. EPSILON) then   ! FOR THE BOUNDARY CONDITION V=0
                VO(J, K-1)=0.                                              ! AT THE LEFT WALL
              end if
            END IF
          END DO
        END DO
 
        CALL VERT_BOTTOM()
        CALL VELOCITYX()
        CALL VELOCITYY()

        IF(M .GT. 2) CALL NON_HYDRO_CGSTAB()

! CONTINUITY ALONG x DIRECTION
        CALL CONTX()

! CONTINUITY ALONG Y DIRECTION
        CALL CONTY()

!UPDATE SEA LEVEL AND VELOCITY
        SLO=SLN
        UO=UN
        VO=VN
        WBO=WBN
        WSO=WSN

! WRITING TO FILE
      !  IF(M.GT.MGEN)THEN
        CALL WRITER_BEACH()
      !  END IF
        DO K=1, KE        
          DO J=1, JE
            IF (ABS(SLO(J, K)) .GT. 80000 .AND. CELL(J, K) .EQ. 1)then  ! terminate program if SLO
              PRINT*, J,K, SLO(J, K)                                    ! is too large                      
              GO TO  203
            END IF
          END DO
        END DO

! JUST FOR QUICK ORIENTATION
        DO k=KS, KE
          DO j=JS, JE
            IF(SLO(J, K) .GT. ZMAX(J, K) .AND. CELL(J, K) .EQ. 1) then
              ZMAX(J, K)=SLO(J, K)
            END IF
            IF(ZMAX(J, K) .GT. SLMA) then
              SLMA=ZMAX(J, K)
              JM=J
              KM=K
            END IF
          END DO
        END DO

! WRITING TO FILE
      !  CALL WRITER_BEACH()

      END DO  ! TIME LOOP
      CALL WRITER_MAX()

 203  CONTINUE

      PRINT*,'SLMAX=',SLMA,'JMAX=',JM, 'KMAX=',KM
 !    END  program SLOSH

    CONTAINS
 !SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS


      SUBROUTINE DEPTH4_RIGHT()

      READ (25) ((H(J,K),J=1,JE), K=1,KE)
      H = -100.*H

! DEPTH AT U POINTS     
      DO K=1, KE
        DO j=js+1, je
          DHU(J, K)=(H(J-1, K)+H(J, K))*0.5
        END DO
      END DO
      DO K=1,KE
        DHU(JS,K) = DHU(JS+1,K)
      END DO

 ! DEPTH AT V POINTS      
      DO K=KS, KE-1
        DO J=1,JE
          DHV(J, K)=(H(J, K+1)+H(J, K))*0.5
        END DO   
      END DO

      DO J=1,JE
        DHV(J, KE)=DHV(J, KE-1)
      END DO

 ! SMOOTH BATHYMETRY  USED IN NON_HYDRO_CGSTAB
      READ (26) ((HF(J,K),J=1,JE),K=1,KE)
      HF = -100.*HF
 
      END SUBROUTINE



      SUBROUTINE TRIGON()
! calculates LATITUDE AND COS SIN
      REAL :: A1, A2, A3, BLAT, RLAT, HALF
  
      HALF=0.5*DF
      BLAT=18.                              ! STARTING LATITUDE IN DEGREES
                                            !  LATITUDE IS NEGATIVE IN SOUTHERN HEMISPHERE
      DO K=KS+1,KE
        RLAT=BLAT+DF*FLOAT(K-1)   ! STEPPING IN 1DEG
        A1=RLAT*RAD
        A2=(RLAT+HALF)*RAD
        A3=(RLAT-HALF)*RAD
        SIZ(K)=SIN(A1)
        SIV(K)=SIN(A2)
        CSZ(K)=COS(A1)
        CSV(K)=COS(A2)
        CSV1(K)=COS(A3)
      END DO
      CSV(Ks)=  CSV(Ks+1)
      CSZ(Ks)=csz(ks+1)
      CSV1(Ks)=  CSV1(Ks+1)

      DO K=KS,KE
        CSVZ(K)=CSV(K)/CSZ(K)
        CSVZ1(K)=CSV1(K)/CSZ(K)
      END DO

      END SUBROUTINE

!====================================


      SUBROUTINE LEFT_FORCING()
      REAL, parameter :: am = 10.
      REAL :: W, FIM
      INTEGER :: LIM,LIMJ

      W=2.*PI/TPF     ! FREQUENCY
               ! AM is AMPLITUDE

      DO K=KS,KE
      !   GAM=SQRT(G*H(JS+1,K))*T/HX
        LIM=M-(KE-K)*3
        IF (LIM .LE. 0) then
          SLN(JS, K)=0.
          GO TO 71
        END IF

        FIM=FLOAT(LIM)
          SLN(JS,K)=am*sin(w*FIM*T)
         ! STEM=am*sin(w*m*T)
         ! SLO(JS,K)=2*SLN(JS,K)-UO(JS+1,K)*SQRT(H(JS+1,K)/G)
         ! SLN(JS,K)=  STEM+( (1-GAM)*SLO(JS,K) +GAM*SLO(JS+1,K))
  71    SLO(JS,K)=SLN(JS,K)
      END DO

      END SUBROUTINE



      SUBROUTINE NORTH_FORCING()
      REAL, parameter :: am = 10.
      REAL :: W, FIMJ
      Integer :: LIMJ

      W=2.*PI/TPF   ! FREQUENCY

      DO j=JS,JE
        LIMJ=M-J*2
        IF(LIMJ .LE. 0) then
          SLN(J, KE)=0.
          GO TO 72
        END IF
        FIMJ=FLOAT(LIMJ)

       ! GAM=SQRT(G*H(JS+1,K))*T/HX
     72 SLN(J,KE)=am*sin(w*FIMJ*T)
        ! STEM=am*sin(w*m*T)
        ! SLO(JS,K)=2*SLN(JS,K)-UO(JS+1,K)*SQRT(H(JS+1,K)/G)
        ! SLN(JS,K)=  STEM+( (1-GAM)*SLO(JS,K) +GAM*SLO(JS+1,K))
          SLO(J, KE)=SLN(J, KE)
      END DO

      END SUBROUTINE




      SUBROUTINE SOUTH_FORCING()
      REAL, parameter :: am = 10.
      REAL :: W
 
      W=2.*PI/TPF   ! FREQUENCY

      DO j=JS,JE
       ! GAM=SQRT(G*H(JS+1,K))*T/HX
        SLN(J, KS)=am*sin(w*m*T)
       !  STEM=am*sin(w*m*T)
       !  SLO(JS,K)=2*SLN(JS,K)-UO(JS+1,K)*SQRT(H(JS+1,K)/G)
       !  SLN(JS,K)=  STEM+( (1-GAM)*SLO(JS,K) +GAM*SLO(JS+1,K))
        SLO(J, KS)=SLN(J, KS)
      END DO
      END SUBROUTINE



      SUBROUTINE RIGHT_FORCING()
      REAL, parameter :: am = 100.
      REAL :: W
 
      W=2.*PI/TPF
      DO K=KS,KE
        SLO(jE, K)=am*sin(w*m*T)
        SLN(JE, K)=SLO(JE, K)
      END DO
      END SUBROUTINE



      SUBROUTINE RIGHT_RADIATING()
      REAL :: HTOT

      DO K=KS+1,KE-1
        HTOT= H(JE-1, K)-ETO(JE-1, K)+SLO(JE-1, K)
        IF (HTOT .LT. 0.) HTOT=0.
        SLO(JE, K)=UO(JE, K)*SQRT(HTOT/G)
        SLN(JE, K)=SLO(JE, K)
      END DO
      END SUBROUTINE



      SUBROUTINE LEFT_RADIATING()
      REAL :: HTOT
              
      DO K=KS,KE
        HTOT= H(JS+1,K)-ETO(JS+1,K)+SLO(JS+1,K)
        IF (HTOT .LT. 0.) HTOT=0.
        SLO(JS, K)=-UO(JS+1, K)*SQRT(HTOT/G)
        SLN(JS, K)=SLO(JS, K)
      END DO
      END SUBROUTINE



      SUBROUTINE NORTH_RADIATING()
      REAL :: HTOT

!Special
      DO j=JS,JE
        HTOT= H(J, KE)-ETO(J, KE)+SLO(J, KE)
        IF (HTOT .LT. 0.) HTOT=0.
        SLO(J, KE)=VO(J, KE-1)*SQRT(HTOT/G)
       !  SLO(J,KE)=   SLO(J,KE-1)
        SLN(J, KE)=SLO(J, KE)
      END DO
      END SUBROUTINE



      SUBROUTINE SOUTH_RADIATING()
      REAL :: HTOT

      DO j=JS,JE
        HTOT= H(J, KS)-ETO(J, KS)+SLO(J, KS)
        IF (HTOT .LT. 0.) HTOT=0.
        SLO(J, KS)=-VO(J, KS)*SQRT(HTOT/G)
        SLN(J, KS)=SLO(J, KS)
      END DO
      END SUBROUTINE




    !SUBROUTINE ENERGY_FLUX
    !DO J=1,JE
    !DO K=1,KE
    !
    !IF (CELL(J,K).EQ.1.AND.ABS(SLO(J,K).GT.0.5) THEN
    !IF(M(J,K).LT.9000)THEN
    !M(J,K)=M(J,K)+1
    !ENX(J,K)=ENx(J,K)+(UO(J,K)+UO(J+1,K))*0.5*SLO(J,K)
    !ENY(J,K)=ENY(J,K)+(VO(J,K-1)+VO(J,K)*0.5*SLO(J,K)
    !END IF
    !
    !END IF
    !END DO
    !END DO

    !END SUBROUTINE

 

      SUBROUTINE VERT_BOTTOM()
      REAL :: UT, UTN, UTP, VT, VTN, VTP

      UT=0.; UTN=0.; UTP=0.; VT=0.; VTN=0.; VTP=0.

      DO K=KS+1,KE-1
        DO J=JS+1,JE-1

      !  IF(CELL(J).EQ.1.AND.CELL(J-1).EQ.1)THEN
        UT=0.5*(UO(J, K)+UO(J+1, K))
        UTP=0.5*(UT+ABS(UT))
        UTN=0.5*(UT-ABS(UT))
   !     UTP1=MAX(0,UT)
   !     UTN1=MIN(0,UT)
        VT=0.5*(VO(J, K)+VO(J, K-1))
        VTP=0.5*(VT+ABS(VT))
        VTN=0.5*(VT-ABS(VT))

        WBN(J, K)=-UTP*(H(J, K)-H(J-1, K))/(HX*CSZ(K))  &
          -UTN*(H(J+1, K)-H(J, K))/(HX*CSZ(K))          &
          -VTP*(H(J, K)-H(J,K-1))/HY                    &
          -VTN*(H(J, K+1)-H(J, K))/HY
        WBN(J, K)=WBN(J, K)*REAL(CELL(J, K))

        END DO
      END DO

      END SUBROUTINE

!IIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIII

      SUBROUTINE VELOCITYX()
      REAL, DIMENSION(1:JE,1:KE) :: DU, UNF
      REAL :: HTL, HTR, DDU, MAG, MAG1, HTOT, UON, UOP, VAU, VAUN, VAUP

 !.       >>>Update   UO with pressure gradient<<<
      DO K=KS+1,KE-1
        DO j = JS+2, JE-1
          IF(CELL(J-1, K) .EQ. 1 .AND. CELL(J, K) .EQ. 1)then
            VAU=0.25*(VO(J, K)+VO(J-1, K)+VO(J-1, K-1)+VO(J, K-1))
            VAUP=0.5*(VAU+ABS(VAU))
            VAUN=0.5*(VAU-ABS(VAU))
            MAG1= UO(J, K)*UO(J, K)+VAU*VAU
            MAG=SQRT(MAG1)
            DU(J, K)=DHU(J, K)+0.5*(SLO(J-1, K)+SLO(J, K))&
             -0.5*(ETO(J-1, K)+ETO(J, K))
            DDU=DU(J, K)
            IF(DU(J, K) .LE. 2.) DDU=2.   ! DEFENSE AGAINST ZERO DENOM. IN BOTTOM FRICTION
         !   UOPM1=0.5*(UO(j-1,K)+ABS(UO(j-1,K)))
         !   UONM1=0.5*(UO(J-1,K)-ABS(UO(J-1,K)))
            UOP=0.5*(UO(j, K)+ABS(UO(j, K)))
            UON=0.5*(UO(J, K)-ABS(UO(J, K)))
         !  UAJ=0.5*(UO(J,K)+UO(J-1,K))
         !  UAJ1=0.5*(UO(J,K)+UO(J+1,K))
         !  UAP=0.5*(UAJ+ABS(UAJ))
         !  UAN=0.5*(UAJ-ABS(UAJ))
         !  UAP1=0.5*(UAJ1+ABS(UAJ1))
         !  UAN1=0.5*(UAJ1-ABS(UAJ1))

            UN(J, K)=UO(J, K)                                                  &
             +FCT*SIZ(K)*VAU                                                   &
             +NL*T*SIZ(K)*VAU*UO(J, K)/(CSZ(K)*R0)                             &
             -THX*(UOP*(UO(J, K)-UO(J-1, K))+UON*(UO(J+1, K)-UO(J, K)))/CSZ(K) &
  ! -THX*0.5*(UAP1*UO(j,K)+UAN1*UO(j+1,K)- UAP*UO(j-1,K)-UAN*UO(j,K))/CSZ(K)   &
             -THY*(VAUP*(UO(J, K)-UO(J, K-1))+VAUN*(UO(J, K+1)-UO(J, K)))      &
             + GTHX*(SLO(j-1, K) - SLO(j, K))/CSZ(K)                           &
             -r*T*UO(j, K)*MAG/DDU
          END IF

        END DO
      END DO

!   BOUNDARY West
      DO K=KS+1,KE
        IF (CELL(JS, K) .EQ. 1 .AND. CELL(JS+1, K) .EQ. 1)then
          DU(JS+1, K)=DHU(JS+1, K)+0.5*(SLO(JS, K)+SLO(JS+1, K))
          IF (DU(JS+1, K) .LE. 5.) DU(JS+1, K)=5.
          UN(JS+1, K)=UO(JS+1, K)-GTHX*(SLO(JS+1, K)-SLO(JS, K))/CSZ(K)        &
           -r*T*UO(JS+1, K)*abs(UO(JS+1, K))/du(JS+1, K)
        END IF
      END DO

!  BOUNDARY EAST
      DO K=KS+1,KE-1
        IF (CELL(JE-1, K) .EQ. 1 .AND. CELL(JE, K) .EQ. 1)then
          DU(JE, K)=DHU(JE, K)+0.5*(SLO(JE-1, K)+SLO(JE, K))
          DDU=DU(JE, K)
          IF (DU(JE, K) .LE. 5.) DDU=5.
          UN(JE, K)=UO(JE, K)+GTHX*(SLO(JE-1, K)-SLO(JE,K))/CSZ(K)             &
           -r*T*UO(JE, K)*abs(UO(JE, K))/DDU
        END IF
      END DO

!  BOUNDARY SOUTH (K=KS)
      DO j = JS+2, JE-1
        IF (CELL(J-1, KS) .EQ. 1 .AND. CELL(J, KS) .EQ. 1)then
          VAU=0.5*(VO(J, KS)+VO(J-1, KS))
          VAUP=0.5*(VAU+ABS(VAU))
          VAUN=0.5*(VAU-ABS(VAU))
          MAG1= UO(J, KS)*UO(J, KS)+VAU*VAU
          MAG=SQRT(MAG1)
          DU(J, KS)=DHU(J, KS)+0.5*(SLO(J-1, KS)+SLO(J, KS))
          DDU=DU(J, KS)
          if (DU(J, KS) .LE. 2.) DDU=2.   ! DEFENSE AGAINST ZERO DENOM. IN BOTTOM FRICTION

          UOP=0.5*(UO(j, KS)+ABS(UO(j, KS)))
          UON=0.5*(UO(J, KS)-ABS(UO(J, KS)))

          UN(J, KS)=UO(J, KS) &
           +NL*T*SIZ(KS)*VAU*UO(J,KS)/(CSZ(KS)*R0)                                &
           -THX*(UOP*(UO(J, KS)-UO(J-1, KS))+UON*(UO(J+1, KS)-UO(J, KS)))/CSZ(KS) &
           -THY*VAUN*(UO(J, KS+1)-UO(J, KS))                                      &
           + GTHX*(SLO(j-1, KS) - SLO(j, KS))/CSZ(KS)                             &
           -r*T*UO(j, KS)*MAG/DDU
        END IF
      END DO

! BOUNDARY NORTH (K=KE)
      DO j = JS+2, JE-1
        IF (CELL(J-1, KE) .EQ. 1 .AND. CELL(J,KE) .EQ. 1)then
          VAU=0.5*(VO(J-1, KE-1)+VO(J, KE-1))
          VAUP=0.5*(VAU+ABS(VAU))
!          VAUN=0.5*(VAU-ABS(VAU))
          MAG1= UO(J, KE)*UO(J, KE)+VAU*VAU
          MAG=SQRT(MAG1)
          DU(J, KE)=DHU(J, KE)+0.5*(SLO(J-1, KE)+SLO(J, KE))

          DDU=DU(J, KE)
          IF(DU(J, KE) .LE. 2.) DDU=2.   ! DEFENSE AGAINST ZERO DENOM. IN BOTTOM FRICTION

          UOP=0.5*(UO(j, KE)+ABS(UO(j, KE)))
          UON=0.5*(UO(J, KE)-ABS(UO(J, KE)))

          UN(J, KE)=UO(J, KE)                                                     &
           +NL*T*SIZ(KE)*VAU*UO(J, KE)/(CSZ(KE)*R0)                               &
           -THX*(UOP*(UO(J, KE)-UO(J-1, KE))+UON*(UO(J+1, KE)-UO(J, KE)))/CSZ(KE) &
           -THY*VAUP*(UO(J, KE)-UO(J, KE-1))                                      &
           + GTHX*(SLO(j-1, KE) - SLO(j, KE))/CSZ(KE)                             &
           -r*T*UO(j, KE)*MAG/DDU
        END IF
      END DO




! FILTER IF NEEDED
!    UNF=UN
!    DO K=2,KE
!      DO j = JS+2, JE - 1
!        UN(J,K)=UNF(J,K)*(1.-ALPHA)+0.25*ALPHA*(UNF(J-1,K)+2.*UNF(J,K)+UNF(J+1,K))
!      END DO
!    END DO

      DO K=KS+1,KE-1
        DO J=JS+1,JE
          IF (CELL(J, K) .EQ. 0 .AND. Cell(j-1, K) .EQ. 0) then
            UN(J, K)=0.
          END IF
        END DO
        DO j = JS+1, JE-1                                                 ! CHANGE 1  TO 2 FOR RADIATION
          IF (Cell(j, K) .EQ. 1 .AND. CELL(J+1, K) .EQ. 0) then
            if ((SLO(j, K) + h(j+1, K)-ETO(J+1, K)) .LE. EPSILON) then    ! FOR THE BOUNDARY CONDITION U=0
              UN(J+1,K)=0.                                                ! AT THE RIGHT WALL
             end if
          END IF
          IF (CELL(J-1, K) .EQ. 0 .AND. CELL(J, K) .EQ. 1) then
            if ((SLO(j, K) + h(j-1, K)-ETO(J-1, K)) .LE. EPSILON) then    ! FOR THE BOUNDARY CONDITION U=0
              UN(J, K)=0.                                                 ! AT THE LEFT WALL
            end if
          END IF
        END DO
      END DO

      RETURN


      DO K=KS+1,KE-1
        DO j = JS+1, JE-1        ! CHANGE 1  TO 2 FOR RADIATION
          IF (Cell(j, K) .EQ. 1 .AND. CELL(J+1, K) .EQ. 0) then
            if ((SLO(j, K) + h(j+1, K)-ETO(J+1, K)) .LE. EPSILON) then   ! FOR THE BOUNDARY CONDITION U=0
              if (HTOT .Le. 10.) HTOT=10.
              UN(J+1, K)=  SLO(J, K)/SQRT(HTOT/G)                        ! AT THE RIGHT WALL
            end if
          END IF

          IF (CELL(J-1, K) .EQ. 0 .AND. CELL(J, K) .EQ. 1) then
            if ((SLO(j, K) + h(j-1, K)-ETO(J-1, K)) .LE. EPSILON) then   ! FOR THE BOUNDARY CONDITION U=0
              HTOT= H(J, K)-ETO(J, K)+SLO(J, K)
              IF(HTOT .Le. 10.) HTOT=10.
              UN(J, K)= - SLO(J, K)/SQRT(HTOT/G)                         ! AT THE LEFT WALL
            end if
          END IF
        END DO
      END DO

      DO K=1,KE
        DO j = JS+1, JE-1
          IF (CELL(J-1, K) .EQ. 1 .AND. CELL(J, K) .EQ. 1 .AND. CELL(J+1, K) .EQ. 0) then
            if ((SLO(J, K) + H(J+1, K)-ETO(J+1, K)) .GT. epsilon) then
              HTOT= H(J, K)-ETO(J, K)+SLO(J, K)
              if (HTOT .Le. 10.) HTOT=10.
              UN(J+1, K)=  SLO(J, K)/SQRT(HTOT/G)
     !         UN(J+1,K)=UN(J,K)
            end if
          END IF
          IF (CELL(J-1, K) .EQ. 0 .AND. CELL(J, K) .EQ. 1 .AND. CELL(J+1, K) .EQ. 1) then
            if ((SLO(J, K) + H(J-1, K)-ETO(J-1, K)) .GT. EPSILON) then
              HTOT= H(J, K)-ETO(J, K)+SLO(J, K)
              if (HTOT .Le. 10.) HTOT=10.
              UN(J,K)= -SLO(J, K)/SQRT(HTOT/G)
           !   UN(J,K)=UN(J+1,K)
            end if
          END IF

 ! WET ONLY IN ONE CELL SURROUNDED BY DRY CELLS.
          IF (CELL(J-1, K) .EQ. 0 .AND. CELL(J, K) .EQ. 1 .AND. CELL(J+1, K) .EQ. 0) then
            HTL= SLO(J, K)+H(J-1, K)-ETO(J-1, K)
            HTR=SLO(J, K)+H(J+1, K)-ETO(J+1, K)
            if (HTL .GT. EPSILON) then
              UN(J, K)= -SQRT(G*HTL)
            end if
            if ((SLO(J, K) + H(J+1, K)-ETO(J+1, K)) .GT. EPSILON) then
              UN(J+1, K)=SQRT(G*HTR)
            end if
          END IF
        END DO
      END DO
   
      END SUBROUTINE

       

      SUBROUTINE CONTX()
      REAL :: SN, SP, ULN, ULP
      REAL, DIMENSION(1:JE,1:KE):: FLUX

      FLUX=0.
      DO K=KS+1,KE-1
        DO J = JS+1, JE   ! SET J=JE WHEN RADIATE, OTHERWISE J=JE-1
          IF (CELL(J, K) .EQ. 1 .AND. CELL(J-1,K) .EQ. 1)then    ! THIS CONTROLS ONLY ONE WALL (LEFT)
            ULP=0.5*(UN(J, K)+ABS(UN(J, K)))
            ULN=0.5*(UN(J, K)-ABS(UN(J, K)))

   !         SLPHX=H(J,K)-H(J-1,K)  !BOTTOM SLOPE
   !         CP= ULP*THX/CSZ(K)
   !         CN= ULN*THX/CSZ(K)
   !	     SDIFX=SLO(J,K)-SLO(J-1,K)
   !         IF(UN(J,K).GT.0)THEN
   !         PSIP=1. !OR 0.
   !
   !         SP=SLO(J-1,K)+0.5*psiP*(1.-2.*CP)*(SLO(J,K)-SLO(J-1,K))
	
   !	     SP=SLO(J-1,K)+(0.5-CP)*SDIFX
             SP=SLO(J-1, K)
   !       END IF
   !       IF(UN(J,K).LE.0.)THEN

   !         PSIN=1. !OR 0.

   !         SN=SLO(J,K)-0.5*PSIN*(1.+2.*CN)*SDIFX
   !       END IF
   !	   SN=SLO(J,K)-(0.5+CN)*SDIFX
          SN=SLO(J,K)

   !       FLUX(J,K)=ULP*(SP-CP*SLPHX)+ULN*(SN-CN*SLPHX)+DHU(J,K)*UN(J,K)&
            FLUX(J, K)=ULP*SP+ULN*SN+DHU(J, K)*UN(J, K)-ULP*ETO(J-1, K)-ULN*ETO(J, K)
          END IF
        END DO
      END DO

      DO K= KS+1,KE-1
        DO J = JS+1, JE-1
          SLNN(J, K)=SLO(J, K)- (FLUX(J+1, K)-FLUX(J, K))*THX/CSZ(K) &
           +ETN(J, K)-ETO(J, K)
        END DO
      END DO
      END SUBROUTINE


!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++

      SUBROUTINE CONTY()
 
      REAL :: SP, SN, VLN, VLP
      REAL, DIMENSION(JS:JE,KS:KE) :: FluxY
      INTEGER:: KEE

      FLUXY=0.

      DO K=KS,KE-1
     ! DO K=KS,KE      
        DO J = JS, JE
                     ! SPECIAL BOUNDARY CONDITION

          SP=0.; SN=0.
          IF (CELL(J, K) .EQ. 1 .AND. CELL(J, K+1) .EQ. 1) then    ! THIS CONTROLS ONLY ONE WALL (LEFT)
            VLP=0.5*(VN(J, K)+ABS(VN(J, K)))
            VLN=0.5*(VN(J, K)-ABS(VN(J, K)))

    !       CP= VLP*THY
    !       CN= VLN*THY

    !       SLPHY=H(J,K+1)-H(J,K)
    !	    SDIFY=SLO(J,K+1)-SLO(J,K)
    !       IF(vn(j,k).gt.0.)THEN
    !         PSIP=1.     !CHARACTERISTICS OR PSIP=0.
    !         sp=slo(j,k)+0.5*PSIP*(1.-2.*CP)*(SLO(J,K+1)-SLO(J,K))
    !	      sp=slo(j,k)+(0.5-CP)*SDIFY
            SP=SLO(J,K)
    !     END IF

    !     IF(vn(j, k) .LE. 0.)THEN

    !        PSIN=1. !CHARACTERISTICS OR PSIN=0.
    !        SN=SLO(J,K+1)-0.5*PSIN*(1.+2.*CN)*(SLO(J,K+1)-SLO(J,K))
    !	     SN=SLO(J,K+1)-(0.5+CN)*SDIFY
           SN=SLO(J,K+1)
    !      END IF

    !      FLUXY(J,K)=VLP*(SP-CP*SLPHY)+VLN*(SN-CN*SLPHY)+DHV(J,K)*VN(J,K) &
           FLUXY(J, K)=VLP*SP+VLN*SN+DHV(J, K)*VN(J, K)                    &
            -VLP*ETO(J, K)-VLN*ETO(J, K+1)
            FLUXY(J, K)=FLUXY(J, K)*CSV(K)

          END IF
        END DO
      END DO
         
! KEE=KE-1
! IF(M .LE. LTIME) THEN
!   DO J = JS+1, JE-1
!    FLUXY(J,KE)=0.

!    END DO
!    KEE=KE
!  END IF

!    DO K = KS+1, KEE
      DO K = KS+1, KE-1          
        DO J = JS+1, JE-1
          SLN(J,K)=SLNN(J,K)- (FLUXY(J,K)-FLUXY(J,K-1))*THY/CSZ(K)

 !        IF ((SLN(J,K)+H(J,K)-ETN(J,K)) .LT. EPSILON) then

 !           SLN(J,K)=-H(J,K)
 !           D(J,K)=SLN(J,K)+H(J,K)
 !           Cell(J,K) = 0
 !        ELSE
 !          Cell(J,K) = 1
 !          D(J,K)=H(J,K)+SLN(J,K)
 !        END IF

        END DO
      END DO
      END SUBROUTINE



      SUBROUTINE VELOCITYY
      REAL, DIMENSION(1:JE,1:KE) :: DV, VNF
      REAL :: VON, VOP, UAV, UAVN, UAVP, MAG, MAG1, DDV, HTOT, SLD, SLU

      DV=0.
      DO j = JS, JE-1
        IF (CELL(J, KS) .EQ. 1 .AND. CELL(J,KS+1) .EQ. 1) THEN
          DV(J,KS)=DHV(J,KS)+0.5*(SLO(J,KS)+SLO(J,KS+1)) &
              - 0.5*(ETO(J,KS)+ETO(J,KS+1))

          if (DV(J,KS) .LE. 5.) DV(J,KS) =5.
          UAV=0.25*(UN(J,KS)+UN(J+1,KS)+UN(J+1,KS+1)+UN(J,KS+1))
          VN(J,KS)=VO(J,KS) &
            -FCT*SIV(K)*UAV&
            -NL*T*UAV*UAV*SIV(K)/(CSV(K)*R0)&
            -GTHY*(SLO(J,KS+1)-SLO(J,KS))-r*T*VO(J,KS)*abs(VO(J,KS))/dV(J,KS)
        END IF
      END DO

! JUST FOR JS
      DO K=KS+1,KE-1
        IF (CELL(JS, K) .EQ. 1 .AND. CELL(JS, K+1) .EQ. 1) THEN
          UAV=0.25*(UN(JS, K)+UN(JS+1, K)+UN(JS+1, K+1)+UN(JS, K+1))
          UAVP=0.5*(UAV+ABS(UAV))
          UAVN=0.5*(UAV-ABS(UAV))
          MAG1=UAV*UAV+VO(JS,K)*VO(JS,K)
          MAG=SQRT(MAG1) 
          DV(J,K)=DHV(J,K)+0.5*(SLO(JS,K)+SLO(JS,K+1))   &
                              -0.5*(ETO(JS,K)+ETO(JS,K+1))
          DDV=DV(JS,K)
          if (DV(JS, K) .LE. 2.) DDV=2.   ! DEFENSE AGAINST ZERO DENOM. IN BOTTOM FRICTION
          VOP=0.5*(VO(jS, K)+ABS(VO(jS, K)))
          VON=0.5*(VO(JS, K)-ABS(VO(JS, K)))
      !    VAK=0.5*(VO(JS,K)+VO(JS,K+1))
      !    VAKP=0.5*(VAK+ABS(VAK))
      !    VAKN=0.5*(VAK-ABS(VAK))
      !    VAK1=0.5*(VO(JS,K)+VO(JS,K-1))
      !    VAK1P=0.5*(VAK1+ABS(VAK1))
      !    VAK1N=0.5*(VAK1-ABS(VAK1))
          VN(jS, K)=VO(JS, K)                                                      &
           -THY*(VOP*(VO(J,K)-VO(J,K-1))+VON*(VO(J,K+1)-VO(J,K)))                  &
      !    -THY*0.5*(VAKP*VO(JS,K)+VAKN*VO(JS,K+1)-VAK1P*VO(JS,K-1)-VAK1N*VO(JS,K))&
           -THX*UAVN*(VO(JS+1,K)-VO(JS,K))/CSV(K)                                  &
           + GTHY*(SLO(jS,K) - SLO(jS,K+1))                                        &
           -FCT*SIV(K)*UAV                                                         &
          -NL*T*UAV*UAV*SIV(K)/(CSV(K)*R0)                                         &
           -r*T*VO(jS,K)*MAG/DDV
         END IF
      END DO


      DO K=KS+1,KE-1
        DO j = JS+1, JE-1
          IF(CELL(J, K) .EQ. 1 .AND. CELL(J, K+1) .EQ. 1)THEN
          UAV=0.25*(UN(J, K)+UN(J+1, K)+UN(J+1, K+1)+UN(J, K+1))
          UAVP=0.5*(UAV+ABS(UAV))
          UAVN=0.5*(UAV-ABS(UAV))
          MAG1=UAV*UAV+VO(J, K)*VO(J, K)
          MAG=SQRT(MAG1)
          DV(J, K)=DHV(J, K)+0.5*(SLO(J, K)+SLO(J, K+1)) &
                              -0.5*(ETO(J, K)+ETO(J, K+1))
          DDV=DV(J,K)
          if (DV(J, K) .LE. 2.) DDV=2.                   ! DEFENSE AGAINST ZERO DENOM. IN BOTTOM FRICTION
          VOP=0.5*(VO(j,K)+ABS(VO(j,K)))
          VON=0.5*(VO(J,K)-ABS(VO(J,K)))
   !       VAK=0.5*(VO(J,K)+VO(J,K+1))
   !       VAKP=0.5*(VAK+ABS(VAK))
   !       VAKN=0.5*(VAK-ABS(VAK))
   !       VAK1=0.5*(VO(J,K)+VO(J,K-1))
   !       VAK1P=0.5*(VAK1+ABS(VAK1))
   !       VAK1N=0.5*(VAK1-ABS(VAK1))
          VN(j, K)=VO(J, K)                                                       &
           -THY*(VOP*(VO(J, K)-VO(J, K-1))+VON*(VO(J, K+1)-VO(J, K)))             &
   !       -THY*0.5*(VAKP*VO(J,K)+VAKN*VO(J,K+1)-VAK1P*VO(J,K-1)-VAK1N*VO(J,K))   &
           -THX*(UAVP*(VO(J, K)-VO(J-1, K))+UAVN*(VO(J+1, K)-VO(J, K)))/CSV(K)    &
            + GTHY*(SLO(j, K) - SLO(j, K+1))                                      &
           -FCT*SIV(K)*UAV                                                        &
           -NL*T*UAV*UAV*SIV(K)/(CSV(K)*R0)                                       &
           -r*T*VO(j,K)*MAG/DDV
          END IF
        END DO
      END DO

! FILTER IF NEEDED
!   VNF=VN
!   DO K=KS+1,KE-1 
!     DO j = JS, JE
!       VN(J,K)=VNF(J,K)*(1.-ALPHA)+0.25*ALPHA*(VNF(J,K-1)+2.*VNF(J,K)+VNF(J,K+1))
!     END DO
!   END DO

      DO J=JS,JE                         ! THIS IS ONLY BEAUTIFICATION
        DO K=KS+1, KE-1                        
          IF (CELL(J, K) .EQ. 0 .AND. CELL(J, K+1) .EQ. 0) THEN
            VN(J, K)=0.
          END IF
        END DO

        DO K = KS+1, KE-1
          IF (Cell(J, K) .EQ. 1 .AND. CELL(J, K+1) .EQ. 0) then
            if ((SLO(J, K) + h(J, K+1)-ETO(J, K+1)) .LE. EPSILON) THEN    ! FOR THE BOUNDARY CONDITION V=0
              Vn(J, K) = 0.                                               ! AT THE RIGHT WALL
            end if
          END IF

          IF (CELL(J, K-1) .EQ. 0 .AND. CELL(J, K) .EQ. 1) THEN
            if ((SLO(J, K) + h(J, K-1)-ETO(J, K-1)) .LE. EPSILON) THEN    ! FOR THE BOUNDARY CONDITION V=0
              VN(J, K-1)=0.                                               ! AT THE LEFT WALL
            end if
          END IF
        END DO
      END DO

      RETURN


      DO K = KS+1, KE-1
        DO J=JS,JE-1
          IF (Cell(J, K) .EQ. 1 .AND. CELL(J, K+1) .EQ. 0) then
            if ((SLO(J, K) + h(J, K+1)-ETO(J, K+1)) .LE. EPSILON) THEN    ! FOR THE BOUNDARY CONDITION V=0
              HTOT= H(J, K)-ETO(J, K)+SLO(J, K)
              if (HTOT .Le. 10.) HTOT=10.
              VN(J, K)=SLO(J, K)/SQRT(HTOT/G)                             ! AT THE RIGHT WALL
            end if
          END IF

          IF (CELL(J, K-1) .EQ. 0 .AND. CELL(J, K) .EQ. 1) THEN
            if ((SLO(J, K) + h(J, K-1)-ETO(J, K-1)) .LE. EPSILON) THEN    ! FOR THE BOUNDARY CONDITION V=0
              HTOT= H(J, K)-ETO(J, K)+SLO(J, K)
              if (HTOT .Le. 10.) HTOT=10.
              VN(J, K-1)=-SLO(J, K)/SQRT(HTOT/G)                          ! AT THE LEFT WALL
            end if
          END IF
        END DO
      END DO

      DO K = KS+1, KE-1
        DO J=1,JE-1
           IF (CELL(J, K-1) .EQ. 1 .AND. Cell(J, K) .EQ. 1 .AND. CELL(J, K+1) .EQ.0) then
             if ((SLO(J, K) + h(J, K+1)-ETO(J, K+1)) .GT. EPSILON) then
               HTOT= H(J, K)-ETO(J, K)+SLO(J, K)
               if (HTOT .Le. 10.)HTOT=10.
               VN(J, K)=SLO(J, K)/SQRT(HTOT/G)
          !     VN(J,K)=VN(J,K-1)
             end if
           END IF

           IF (CELL(J, K-1) .EQ. 0 .AND. CELL(J, K) .EQ. 1 .AND. CELL(J, K+1) .EQ. 1) THEN
             if ((SLO(J, K) + h(J, K-1)-ETO(J, K-1)) .GT. EPSILON) then
               HTOT= H(J, K)-ETO(J, K)+SLO(J, K)
               if (HTOT .Le. 10.) HTOT=10.
               VN(J,K-1)=-SLO(J, K)/SQRT(HTOT/G) 
         !      VN(J,K-1)=VN(J,K)
             end if
           END IF

 ! LONELY WET CELLS
           IF (CELL(J, K-1) .EQ. 0 .AND. CELL(J, K) .EQ. 1 .AND. CELL(J, K+1) .EQ. 0) THEN
             SLD= SLO(J, K) + h(J, K-1)-ETO(J, K-1)
             SLU=SLO(J, K) + h(J, K+1) -ETO(J, K+1)
             if ((SLO(J, K) + h(J, K-1)-ETO(J, K-1)) .GT. EPSILON) then
               VN(J,K-1)=-SQRT(G*SLD)
             end if
             if ((SLO(J, K) + H(J, K+1)-ETO(J, K+1)) .GT. EPSILON) then
               VN(J,K)=SQRT(G*SLU)
             end if
           END IF
        END DO
      END DO

      END SUBROUTINE


!++++++++++++++++++++++++++++++++++++++++++++++++++++

!     Last change:  YY    4 Dec 2007    9:41 am
!------------------------------------------------------------------------
 
!++++++++++++++++++++++++++++++++++++++++++++++++++++

!     Last change:  YY    4 Dec 2007    9:41 am
!------------------------------------------------------------------------
 
SUBROUTINE NON_HYDRO_CGSTAB
!------------------------------------------------------------------------
!       Non-hydrostatic Pressure solver in Spherical coordinates
!       with CGSTAB solver wtih NEW WET-DRY Neuman B.C.
!------------------------------------------------------------------------
!
!  This program calculates coefficients of the linear equation system
!  to obtain q using the conjugate gradient squred stabilized solver
!  for 5-diagonal, and both non-symmetric and symmetric matrices.
!  (subroutine CGSTAB2D) with Dirichelt B.C. at WET-DRY boundaries.
!
!   originated by  Z.Kowalik  2006  IMS, University of Alaska Fairbanks
!   written by     Y.Yamazaki 2007  ORE, University of Hawaii Manoa
!------------------------------------------------------------------------
   !   USE ALLSUB
     !IMPLICIT NONE

      REAL(KIND=DP), PARAMETER:: rho     = 1.0_DP       ! rho(g/cm^3) : water density
     !REAL(KIND=DP), PARAMETER:: rho     = 1.0d3       ! rho(kg/m^3) : water density

      INTEGER,  PARAMETER:: NJK  = (je-js+1)*(ke-ks+1)

      REAL(KIND=DP), DIMENSION( js:je, ks:ke ):: Ax, Ay,DLIM,A1q,B1Q
      REAL(KIND=DP), DIMENSION( js:je, ks:ke ):: qf    ! temp q for filtering
      REAL(KIND=DP), DIMENSION( 1:NJK )       :: Phi   ! = q
      REAL(KIND=DP), DIMENSION( 1:NJK )       :: AL, AR, AT, AB, AP, QP
      REAL(KIND=DP), DIMENSION( 1:NJK )       :: SA
      REAL(KIND=DP):: RES0
      REAL(KIND=DP):: VOL, QUO

      INTEGER,  DIMENSION(1:je) :: LJ
      INTEGER :: jk

     ! Phi = NJK*0.0_DP
      A1q = NJK*0.0_DP; B1q = NJK*0.0_DP

      AR = NJK*0.0_DP; AL = NJK*0.0_DP
      AT = NJK*0.0_DP; AB = NJK*0.0_DP
      AP = NJK*0.0_DP; QP = NJK*0.0_DP

      qf = 0.0_DP
 !     q  = 0.0_DP
      DLIM=0.0_DP
      a1Q=0.0_dp
      B1q=0.0_DP

!.. convering 2D Array into 1D !
      DO j = js, je
        LJ(j) = (j-1)*(ke-ks+1)
      END DO
!.. converting 1D Array into 2D
!SETTING Q=0 IN THE DRY CELLS
      DO k = ks, ke
        DO j = js, je
           jk = LJ(j) + k
           PHI(JK)=Q(J,K)
        END DO
      END DO
! FOR THE LINE JS SET OPEN BOUNDARY BY SETTING Q(JS)=0
!  DO k = ks, ke
!           jk =   k
!    q( js, k) = 0._DP
!PHI(JK)= q( js, k)
!END DO

!------------------------------------------------------------------------------
!  Limited Total Water Depth, Dlim
!------------------------------------------------------------------------------
      DO k = ks, ke
        DO j = js, je
           Dlim(j,k)=  HF(j,k) +slo(j,k)           ! try both !!
           !Dlim(j,k)=  H(j,k) +slo(j,k) -eto(j,k)
           IF ( ABS(Dlim(j,k)) .LT. 0.01_DP ) THEN
              Dlim(j,k)=0.01_DP  !(cm) limit D for q cal (D =10 cm)
           END IF
        END DO
      END DO

!------------------------------------------------------------------------------
!  A1q,Coefficient (A1 in Kowalik 2006)
!------------------------------------------------------------------------------
      DO k = ks, ke
        DO j = js+1, je
          IF (CELL( j ,k) .EQ. 1       &   ! *** Yoshiki 1202 corrected
                          .AND.        &   ! *** Yoshiki 1202 corrected
              CELL(j-1,k) .EQ. 1 ) THEN    ! *** Yoshiki 1202 corrected
             A1q(j,k) = (                                       &
                          ( slo(j  ,k) -HF(j  ,k) +eto(j  ,k) )  &
                         -( slo(j-1,k) -HF(j-1,k) +eto(j-1,k) )  &
                         )                                      &
                          / ( Dlim(j,k) +Dlim(j-1,k) )
          END IF                           ! *** Yoshiki 1202 corrected
        END DO
      END DO

!.. For LEFT Boundary (j = js)

      DO k = ks, ke
        IF (CELL( js ,k) .EQ. 1 ) THEN     ! *** Yoshiki 1202 corrected
          A1q(js,k) = A1q(js+1,k)
        END IF                             ! *** Yoshiki 1202 corrected
      END DO


!------------------------------------------------------------------------------
!  B1q, Coefficient (B1 in Kowalik 2006)
!------------------------------------------------------------------------------
      DO k = ks, ke-1
        DO j = js, je
          IF (CELL(j, k ) .EQ. 1       &   ! *** Yoshiki 1202 corrected
                          .AND.        &   ! *** Yoshiki 1202 corrected
              CELL(j,k+1) .EQ. 1 ) THEN    ! *** Yoshiki 1202 corrected
             B1q(j,k) = (                                       &
                          ( slo(j,k+1) -HF(j,k+1) +eto(j,k+1) )  &
                         -( slo(j,k)   -HF(j,k)   +eto(j,k)   )  &
                         )                                      &
                          / ( Dlim(j,k+1) +Dlim(j,k) )
          END IF                           ! *** Yoshiki 1202 corrected
        END DO
      END DO

!.. For TOP Boundary (k = ke-1)
      DO j = js, je
        IF (CELL( j ,ke) .EQ. 1 ) THEN     ! *** Yoshiki 1202 corrected
             B1q(j,ke) = B1q(j,ke-1)
        END IF                             ! *** Yoshiki 1202 corrected
      END DO

!------------------------------------------------------------------------------
!  AL,AR,AT,AB,AP,QP Coefficients of Linear Equation System
!------------------------------------------------------------------------------
      DO k = ks, ke
        DO j = js, je
             Ax(j,k)=(hx*csz(k))**2/(0.5*t)
             Ay(j,k)=( hx*csz(k)/hy )**2
        END DO
      END DO

      DO k = ks+1, ke-1
        DO j = js+1, je-1

          jk = LJ(j) + k

          AL(jk) = -1._DP +A1q(j,k)


          AR(jk) = -1._DP -A1q(j+1,k)


          AT(jk) = Ay(j,k)  *( csv( k ) /csz(k) ) *( -1._DP -B1q(j,k))


          AB(jk) = Ay(j,k) *( csv(k-1) /csz(k) ) *( -1._DP +B1q(j,k-1) )


          AP(jk) = (  (1._DP +A1q(j,k))  +(1._DP-A1q(j+1,k)  )) &
                      +(Ay(j,k)/csz(k)) * (                            &
                                        ( 1._DP -B1q(j, k ) )*csv( k )    &
                                       +( 1._DP +B1q(j,k-1) )*csv(k-1)  ) &
                    +4._DP*(hx*csz(k))**2/(Dlim(j,k)**2)


          QP(jk) = rho*Ax(j,k)*( -(          un(j+1,k)           -un(j, k ) ) / (hx*csz(k))       &
                                 -(   csv(k)*vn( j ,k)  -csv(k-1)*vn(j,k-1) ) / (hy*csz(k))       &
                                 -( wso(j,k) + wbo(j,k) -2.*wbn(j,k) )/Dlim(j,k)                  &
                                                                                                 )

        END DO
      END DO
      !THE BOUNDARY CONDITIONS AT THE TOP AND BOTTOM REQUIRE
      !THAT WET DOMAIN  BE DISCONECTED FROM DRY DOMAIN
      !FIRST VERTICAL BOUNDARIES
         DO k = ks+1, ke-1
        DO j = js+1, je-1
        IF(CELL(J,K)==1.AND. CELL(J,K+1)==0)THEN

          jk = LJ(j) + k
          AP(JK)=AP(JK)+AT(JK)
          AT(JK)=0._DP
          END IF
          END DO
          END DO
             DO k = ks+1, ke-1
        DO j = js+1, je-1
  !      IF(CELL(J,K-1)==0.AND. CELL(J,K)==1)THEN  correction
           IF(CELL(J,K)==0.AND. CELL(J,K+1)==1)THEN
          jk = LJ(j) + k
          AP(JK+1)=AP(JK+1)+AB(JK+1)  !correction
          AB(JK+1)=0._DP              !correction
          END IF
          END DO
          END DO
       !the same approach for the horizontal boundaries
          DO k = ks+1, ke-1
        DO j = js+1, je-1
        IF(CELL(J-1,K)==0.AND. CELL(J,K)==1)THEN

          jk = LJ(j) + k
          AP(JK)=AP(JK)+AL(JK)
          AL(JK)=0._DP
          END IF
          END DO
          END DO
             DO k = ks+1, ke-1
        DO j = js+1, je-1
        IF(CELL(J,K)==0.AND. CELL(J-1,K)==1)THEN  !correction

          jk = LJ(j-1) + k                 !correction
          AP(JK)=AP(JK)+AR(JK)
          AR(JK)=0._DP
          END IF
          END DO
          END DO

!------------------------------------------------------------------------------
!  Neuman B.C. for Computational Domain Boundary
!------------------------------------------------------------------------------
 !lEFT DRY 
 !         j = js+1
 !        DO k = ks+1, ke-1
 !          jk = LJ(j) + k

 !          AP(jk) = AP(jk) + AL(jk)
 !          AL(jk) = 0.0_DP

 !        ENDDO

!..    LEFT Boundary IS PARTLY DRY
         j = js+1
         DO k = ks+1, ke-1
           jk = LJ(j) + k
         IF(CELL(J,K)==0)THEN

           AP(jk) = AP(jk) + AL(jk)
           AL(jk) = 0.0_DP
           END IF
         ENDDO


!        AND PARTLY WET
          j = js+1
         DO k = ks+1, ke-1
          jk = LJ(j) + k

         IF(CELL(J,K)==1)THEN

          AP(jk) = AP(jk) -1._dp +A1q(js+1,k) ! -1 or +1. slow instability
           AL(jk) = 0._dp   ! -1. doesnot work
           END IF
         ENDDO

!           RIGHT Boundary IS PARTLY DRY
         j = jE-1
         DO k = ks+1, ke-1
           jk = LJ(j) + k
         IF(CELL(J,K)==0)THEN

           AP(jk) = AP(jk) + AR(jk)
           AR(jk) = 0.0_DP
           END IF
         ENDDO
!	AND PARTLY WET
           j = jE-1
         DO k = ks+1, ke-1
           jk = LJ(j) + k
         IF(CELL(J,K)==1)THEN

           AP(jk) = AP(jk)   -1._DP -A1q(j+1,k)
           AR(jk) = 0.0_DP
           END IF
         ENDDO



!..    RIGHT Boundary
!         j = je-1
!         DO k = ks+1, ke-1
!           jk = LJ(j) + k
!           AP(jk) = AP(jk) + AR(jk)
!           AR(jk) = 0.0_DP
!         ENDDO

!..    BOTTOM Boundary IS PARTLY DRY
         k = ks+1
         DO j = js+1, je-1
          jk = LJ(j) + k
 	IF(CELL(J,K)==0)THEN
           AP(jk) = AP(jk) + AB(jk)
           AB(jk) = 0.0_DP
	END IF
         ENDDO
!	AND PARTLY WET 
	k = ks+1
         DO j = js+1, je-1
          jk = LJ(j) + k
 	IF(CELL(J,K)==1)THEN
           AP(jk) = AP(jk)+ Ay(j,k) *( csv(k-1) /csz(k) ) *( -1._DP +B1q(j,k-1))
           AB(jk) = 0.0_DP
	END IF
         ENDDO
!..    TOP Boundary IS PARTLY DRY
         k = ke-1
         DO j = js+1, je-1
           jk = LJ(j) + k
	IF(CELL(J,K)==0)THEN
           AP(jk) = AP(jk) + AT(jk)
           AT(jk) = 0.0_DP
	END IF
        ENDDO
! AND PARTLY WET

k = ke-1
         DO j = js+1, je-1
           jk = LJ(j) + k
	IF(CELL(J,K)==1)THEN
           AP(jk) = AP(jk) + Ay(j,k)  *( csv( k ) /csz(k) ) *( -1._DP -B1q(j,k))
           AT(jk) = 0.0_DP
	END IF
        ENDDO

!**********************************************
! This is the new BC I (NOW I IS ZYGMUNT) added

!------------------------------------------------------------------------------
!  Neuman B.C for WET-DRY interface
!------------------------------------------------------------------------------

      DO k = ks+1, ke-1
       DO j = js+1, je-1
          IF (CELL( j ,k) .EQ. 0) THEN
            jk = LJ(j) + k
            AL(jk) = -1._DP
           AR(jk) = -1._DP
           AT(JK)=-1._DP
           AB(JK)=-1._DP
           AP(JK)=4._DP
!            AT(jk) = -Ay(j,k)*( csv( k ) /csz(k) )
!            AB(jk) = -Ay(j,k)*( csv(k-1) /csz(k) )
!            AP(jk) = 2.0+( Ay(j,k) / csz(k) )*( csv(k) +csv(k-1)  )! &
!                     +4.*(hx*csz(k))**2/(Dlim(j,k)**2)
            QP(jk) = 0.0_DP
          END IF
        END DO
      END DO
!=========================================================
! AND NOW DISCONNECT DRY FROM WET (THIS STEP IS UNNECESSARY)
!IT IS JUST FOR 2 PURPOSES. sIMPLER COMPUTATIONS IN dRY DOMAIN
! AND Q WILL BE ZERO IN DRY DOMAIN
!=========================================================
!fIRST ALONG VERTICAL DIRECTION
   DO k = ks+1, ke-1
       DO j = js+1, je-1
          IF (CELL( j ,k) ==0.AND.CELL(J,K+1)==1) THEN
           jk = LJ(j) + k
          AP(JK)=3._DP
          AT(JK)=0._DP
          END IF
          END DO
          END DO

   DO k = ks+1, ke-1
       DO j = js+1, je-1
          IF (CELL( j ,k) ==0.AND.CELL(J,K-1)==1) THEN
           jk = LJ(j) + k
          AP(JK)=3._DP
          AB(JK)=0._DP
          END IF
          END DO
          END DO
!ALONG HORIZONTAL DIRECTION
          DO k = ks+1, ke-1
       DO j = js+1, je-1
          IF (CELL( j ,k) ==0.AND.CELL(J+1,K)==1) THEN
           jk = LJ(j) + k
          AP(JK)=3._DP
          AR(JK)=0._DP
          END IF
          END DO
          END DO

          DO k = ks+1, ke-1
       DO j = js+1, je-1
          IF (CELL( j ,k) ==0.AND.CELL(J-1,K)==1) THEN
           jk = LJ(j) + k
          AP(JK)=3._DP
          AL(JK)=0._DP
          END IF
          END DO
          END DO

!**********************************************


!      IF ( MOD(m,mout) .EQ. 0 ) THEN

!.. Stability, SA
!           DO k = ks+1, ke-1
!             DO j = js+1, je-1
!               jk = LJ(j) + k
!               SA(jk) = (AL(jk)+ AR(jk) +AT(jk) +AB(jk) ) / AP(jk)
!             ENDDO
!           ENDDO

!           WRITE (72) ( (A1q(j,k), j=js,je ), k=ks,ke )
!           WRITE (73) ( (B1q(j,k), j=js,je ), k=ks,ke )
!           WRITE (84) ( (Dlim(j,k), j=js,je ), k=ks,ke )
!
!           WRITE (74) ( ( AL(LJ(j)+k), j = js,je ), k = ks,ke )
!           WRITE (75) ( ( AR(LJ(j)+k), j = js,je ), k = ks,ke )
!           WRITE (76) ( ( AT(LJ(j)+k), j = js,je ), k = ks,ke )
!           WRITE (77) ( ( AB(LJ(j)+k), j = js,je ), k = ks,ke )
!           WRITE (78) ( ( AP(LJ(j)+k), j = js,je ), k = ks,ke )
!           WRITE (79) ( ( QP(LJ(j)+k), j = js,je ), k = ks,ke )
!           WRITE (81) ( ( SA(LJ(j)+k), j = js,je ), k = ks,ke )
!      END IF

!--------------------------------------------------------------
!  Conjugate Gradient Squared Stabilized solver
!  solve q (=Phi) using CGSTAB solver
!--------------------------------------------------------------
      CALL CGSTAB2D(AL, AR, AT, AB, AP, QP, LJ, Phi)

!      IF ( MOD(m,mout) .EQ. 0 ) THEN
!           WRITE (80) ( ( Phi(LJ(j)+k), j = js,je ), k = ks,ke )
!      END IF

!.. convering 1D Array into 2D
      DO k = ks, ke
        DO j = js, je
           jk = LJ(j) + k
             q(j,k) = Phi(jk)
        END DO
      END DO

     ! q( js+1, ks+1:ke-1 ) = q( js+2, ks+1:ke-1 ) !second derivative at the open
      q( js, ks+1:ke-1 ) =q( js+1, ks+1:ke-1 )  ! boundary disappear
      q( je, ks+1:ke-1 ) = q( je-1, ks+1:ke-1 )
      q( js:je, ks ) = q( js:je, ks+1 ) ! cover EAST corner
      q( js:je, ke ) = q( js:je, ke-1 ) ! cover EAST corner

    !   IF(M==24000)THEN
    !   DO K=KS,Ke
    !    write (71,*) (q(J,K),J=1,1420)
    !   END DO
    !   END IF
    !
    !  if(mod(m,1000).eq.0)then
    !  WRITE (72)  (q(j,5),j=1,3000)
    !  END IF

!--------------------------------------------------------------
! Space Filtering of q
!--------------------------------------------------------------
!      qf(js:je,ks:ke)=q(js:je,ks:ke)
!
!      DO k = ks+1, ke-1
!        DO j = js+1, je-1
!!.. Kowalik and Murty 1993 p.145 Eq.3.135 (work!)
!          q(j,k) = qf(j,k) +0.5*ALPHA*(1.-ALPHA)                        &
!                   *(qf(j-1,k)+qf(j+1,k)+qf(j,k-1)+qf(j,k+1)-4*qf(j,k)) &
!                   +0.25*ALPHA**2*(qf(j+1,k-1)+qf(j-1,k-1)              &
!                   +qf(j-1,k+1)+qf(j+1,k+1)-4*qf(j,k))
!!.. Kowalik and Murty 1993 p.147 Eq.3.145
!         !q(j,k) = ALPHA*qf(j,k)                               &
!         !        +0.25*(1.-ALPHA)*(  qf(j-1,k)  +qf(j+1,k)    &
!         !                           +qf(j,k-1)  +qf(j,k+1)    &
!         !                                     -4*qf(j,k)   )
!        END DO
!      END DO
!
!
!      q( js, ks+1:ke-1 ) = q( js+1, ks+1:ke-1 )
!      q( je, ks+1:ke-1 ) = q( je-1, ks+1:ke-1 )
!      q( js:je, ks ) = q( js:je, ks+1 ) ! cover EAST corner
!      q( js:je, ke ) = q( js:je, ke-1 ) ! cover EAST corner





!VERTICAL VELOCITY AT THE SURFACE
DO J=JS+1,JE
DO K=KS,KE
  IF(CELL(J,K).EQ.1)THEN
 WSN(J,K)=WSO(J,K)+WBO(J,K)-WBN(J,K)+2.*T*Q(J,K)/DLIM(J,K)
 END IF
 END DO
END DO
! ANEW STEP IN VELOCITY FROM M+1/2 TO M+1




DO J=JS+1,JE
DO K= KS,KE

 IF(CELL(J,K).EQ.1.AND.CELL(J-1,K).EQ.1)THEN
  UN(J,K)=UN(J,K)-(Q(J,K)+Q(J-1,K))*A1q(J,K)*THX/(2.*CSZ(K)) -&
           (Q(J,K)-Q(J-1,K))*THX/(2.*CSZ(K))
           END IF
END DO
END DO

!UF=UN
!UF(JS,KS:KE)=UN(JS+1,KS:KE)
!DO J=JS+2,JS+12   ! Only A FEW LINES CLOSE TO BOUNDARY
!DO K=KS+1,KE-1
!IF(CELL(J-1,K).EQ.1.AND.CELL(J,K).EQ.1.AND.CELL(J+1,K).EQ.1)THEN
! UN(J,K)=UF(J,K)*(1.-ALPHA1)+0.25*ALPHA1*(UF(J-1,K)&
!          +2.*UF(J,K)+UF(J+1,K))
!          END IF
!          END DO
!          END DO
UN(JS,KS:KE)=UN(JS+1,KS:KE)
DO J=JS+1,JE
DO K= KS,KE-1

   IF(CELL(J,K).EQ.1.AND.CELL(J,K+1).EQ.1)THEN
  VN(J,K)=VN(J,K)-(Q(J,K)+Q(J,K+1))*B1q(J,K)*THY/2.-&
        (Q(J,K+1)-Q(J,K))*THY/2.
        END IF
END DO
END DO
!VN(JS:JE,KE)=VN(JS:JE,KE-1)

! BOUNDARY CONDITIONS FOR VELOCITY
DO K=KS+1,KE-1

          DO J=JS+1,JE
             IF (CELL(J,K).EQ.0.AND.Cell(j-1,K).EQ. 0)THEN
             UN(J,K)=0.

             END IF
             END DO


              DO j = JS+1, JE - 1        ! CHANGE 1  TO 2 FOR RADIATION
      IF (Cell(j,K).EQ.1.AND.CELL(J+1,K).EQ.0) then
       IF ((SLO(j,K) + h(j+1,K)-ETO(J+1,K)).LE.EPSILON) THEN    ! FOR THE BOUNDARY CONDITION U=0
                  UN(J+1,K)=0.                      ! AT THE RIGHT WALL
                  END IF
            END IF

             IF (CELL(J-1,K).EQ.0.AND. CELL(J,K).EQ.1) THEN
    IF ((SLO(j,K) + h(j-1,K)-ETO(J-1,K)).LE.EPSILON) THEN    ! FOR THE BOUNDARY CONDITION U=0
                  UN(J,K)=0.                      ! AT THE LEFT WALL
              END IF
               END IF
        END DO
        END DO

          DO J=JS,JE  ! THIS IS ONLY BEAUTIFICATION
             DO K=KS+1,KE-1
             IF (CELL(J,K).EQ.0.AND. CELL(J,K+1) .EQ. 0)THEN
             VN(J,K)=0.
             END IF
             END DO



              DO K = KS+1, KE - 1
      IF (Cell(J,K).EQ.1.AND.CELL(J,K+1).EQ.0) then
        IF ((SLO(J,K) + h(J,K+1)-ETO(J,K+1)).LE.EPSILON) THEN    ! FOR THE BOUNDARY CONDITION V=0
                              ! AT THE RIGHT WALL
                   VN(J,K)=0.
                  END IF
            END IF

             IF (CELL(J,K-1).EQ.0.AND. CELL(J,K).EQ.1) THEN
           IF ((SLO(J,K) + h(J,K-1)-ETO(J,K-1)).LE.EPSILON) THEN    ! FOR THE BOUNDARY CONDITION V=0

                VN(J,K-1)=0.                      ! AT THE LEFT WALL
               END IF
              end if

END DO
END DO
204 CONTINUE
END SUBROUTINE

!     Last change:  YY    9 Sep 2007    3:27 am
!------------------------------------------------------------------------
      SUBROUTINE CGSTAB2D(AL, AR, AT, AB, AP, QP, LJ, Phi)
!------------------------------------------------------------------------
!    Conjugate Gradient Squared Stabilized solver
!                      for both Non-symmetric and Symmetric Matrices
!------------------------------------------------------------------------
!
!  This program calculates Phi of the linear equation system using the
!  conjugate gradient squred stabilized algorithm (CGSTAB) for 5-diagonal,
!  and both non-symmetric and symmetric matrices.
!
!   The original code was written by
!        Samir Muzaferija, Institut fuer Schiffbau, Hamburg, 1995.
!   (See Ferzinger and Peric (2002) 5.3.6, P.111).
!
!    2007 Y. Yamazaki    ORE, University of Hawaii at Manoa
!------------------------------------------------------------------------
      INTEGER,  PARAMETER:: NJK    = (je-js+1)*(ke-ks+1) ! = je*ke

      REAL(KIND=DP), DIMENSION(1:NJK):: Phi
      REAL(KIND=DP), DIMENSION(1:NJK):: AL, AR, AB, AT, AP, QP
      REAL(KIND=DP), DIMENSION(1:NJK):: RES, RESO
      REAL(KIND=DP), DIMENSION(1:NJK):: PK, ZK, D1
      REAL(KIND=DP), DIMENSION(1:NJK):: UK, VK

      INTEGER, DIMENSION(1:je):: LJ

      REAL(KIND=DP):: RESZERO
      REAL(KIND=DP):: BET, ALF, RESN, RSM              ! new in CGSTAB
      REAL(KIND=DP):: BETO, GAM, OM, UKRESO, VRES, VV  ! new in CGSTAB
     !REAL(KIND=DP):: S0, SK, BET, S2, ALF, RESN, RSM  ! iccg

      INTEGER :: jk, N


!.. Setup Paramters
!------------------------------------------------------------------------

   ! Maximum number of allowed Iteration
      INTEGER,  PARAMETER:: MAXIT  = 2000
   ! Normalized Residuel Sum
      REAL(KIND=DP), PARAMETER:: RESMAX = 0.001_DP
!-----------------------------------------------------------------------


!... calculate the initial residual vector, RESZERO

      RESZERO=0.0_DP
    !     PRINT*,"PHI(6)",PHI(6)
      DO j = js+1, je-1
        DO jk = LJ(j)+ks+1, LJ(j)+ke-1
            RES(jk) = QP(jk)             &
                     -AP(jk)*Phi(jk)     &
                     -AR(jk)*Phi(jk+ke)  &
                     -AL(jk)*Phi(jk-ke)  &
                     -AT(jk)*Phi(jk+1)   &
                     -AB(jk)*Phi(jk-1)

            RESZERO = RESZERO +ABS(RES(jk))
        END DO
      END DO
   !   PRINT*, MAX(RES(JK))
  ! PRINT*,"MAXVALUE= ", MAXVAL(RES)
 ! PRINT*,"MAX_LOCATION= ", MAXLOC(RES)


!... preconditioning the matrix diagonal, D1

      DO j = js+1, je-1
        DO jk = LJ(j)+ks+1, LJ(j)+ke-1
          D1(jk) = 1. / ( AP(jk)                       &
                         -AL(jk)*D1(jk-ke)*AR(jk-ke)   &
                         -AB(jk)*D1(jk-1)*AT(jk-1)   )
        END DO
      END DO


!... initizlize the working arreys and constants

      DO j = js+1, je-1
        DO jk = LJ(j)+ks+1, LJ(j)+ke-1
            RESO(jk) = RES(jk)
            PK(jk)   = 0.0_DP
            UK(jk)   = 0.0_DP
            ZK(jk)   = 0.0_DP
            VK(jk)   = 0.0_DP
        END DO
      END DO

      ALF  =1.0_DP
      BETO =1.0_DP
      GAM  =1.0_DP

!------------------------------------------------------------------------
! START ITERATION
!------------------------------------------------------------------------
      DO N=1,MAXIT

!... calculate beta BET and omega OM

        BET = 0.0_DP

        DO j = js+1, je-1
          DO jk = LJ(j)+ks+1, LJ(j)+ke-1
            BET = BET +RES(jk)*RESO(jk)
          END DO
        END DO

        OM   = BET*GAM / ( ALF*BETO +1.0D-20 )

        BETO = BET

!... calculate PK

        DO j = js+1, je-1
          DO jk = LJ(j)+ks+1, LJ(j)+ke-1
            PK(jk)=RES(jk)+OM*(PK(jk)-ALF*UK(jk))
          END DO
        END DO


!... solve M ZK = PK
!     FORWARD substitution
        DO j = js+1, je-1
          DO jk = LJ(j)+ks+1, LJ(j)+ke-1
            ZK(jk) = (                            &
                       PK(jk)                     &
                       -AL(jk)*ZK(jk-ke)          &
                       -AB(jk)*ZK(jk-1)           &
                                         )*D1(jk)
          END DO
        END DO

        DO j = js+1, je-1
          DO jk = LJ(j)+ks+1, LJ(j)+ke-1
            ZK(jk) = ZK(jk) / ( D1(jk) +1.0D-20 )
          END DO
        END DO

!... backward substitution
        DO j =je-1, js+1, -1
          DO jk = LJ(j)+ke-1, LJ(j)+ks+1, -1
            ZK(jk) = (                            &
                       ZK(jk)                     &
                       -AR(jk)*ZK(jk+ke)          &
                       -AT(jk)*ZK(jk+1)           &
                                         )*D1(jk)
          END DO
        END DO

!... calculate UK = A PK
        DO j = js+1, je-1
          DO jk = LJ(j)+ks+1, LJ(j)+ke-1
            UK(jk) = AP(jk)*ZK(jk)     &
                    +AR(jk)*ZK(jk+ke)  &
                    +AL(jk)*ZK(jk-ke)  &
                    +AT(jk)*ZK(jk+1)   &
                    +AB(jk)*ZK(jk-1)
          END DO
        END DO


!... calculate scalar product UK RESO and GAMMA
        UKRESO = 0.0_DP

        DO j = js+1, je-1
          DO jk = LJ(j)+ks+1, LJ(j)+ke-1
            UKRESO = UKRESO +UK(jk)*RESO(jk)
          END DO
        END DO

        GAM = BET / UKRESO


!... update Phi and calculate w (overwrite RES; it is RES-UPDATE)
        DO j = js+1, je-1
          DO jk = LJ(j)+ks+1, LJ(j)+ke-1
            Phi(jk) = Phi(jk) +GAM*ZK(jk)
            RES(jk) = RES(jk) -GAM*UK(jk)
          END DO
        END DO

!... solve M Y = W
!      overwrite ZK
!      forward substitution
        DO j = js+1, je-1
          DO jk = LJ(j)+ks+1, LJ(j)+ke-1
            ZK(jk) = (                            &
                       RES(jk)                    &
                       -AL(jk)*ZK(jk-ke)          &
                       -AB(jk)*ZK(jk-1)           &
                                         )*D1(jk)
          END DO
        END DO

        DO j = js+1, je-1
          DO jk = LJ(j)+ks+1, LJ(j)+ke-1
            ZK(jk)=ZK(jk)/(D1(jk)+1.0D-20)
          END DO
        END DO


!... backward substitution
        DO j =je-1, js+1, -1
          DO jk = LJ(j)+ke-1, LJ(j)+ks+1, -1
            ZK(jk) = (                            &
                       ZK(jk)                     &
                       -AR(jk)*ZK(jk+ke)          &
                       -AT(jk)*ZK(jk+1)           &
                                         )*D1(jk)
          END DO
        END DO


!... calculate V = A Y (VK = A ZK)
        DO j = js+1, je-1
          DO jk = LJ(j)+ks+1, LJ(j)+ke-1
            VK(jk) = AP(jk)*ZK(jk)      &
                    +AR(jk)*ZK(jk+ke)   &
                    +AL(jk)*ZK(jk-ke)   &
                    +AT(jk)*ZK(jk+1)    &
                    +AB(jk)*ZK(jk-1)
          END DO
        END DO


!... calculate ALPHA (ALF)
        VRES=0.

        VV=0.

        DO j = js+1, je-1
          DO jk = LJ(j)+ks+1, LJ(j)+ke-1
            VRES = VRES +VK(jk)*RES(jk)
            VV   = VV   +VK(jk)*VK(jk)
          END DO
        END DO

        ALF = VRES /( VV +1.0D-20 )


!... update Phi and residual (RES) vectors
        RESN=0.

        DO j = js+1, je-1
          DO jk = LJ(j)+ks+1, LJ(j)+ke-1
            Phi(jk)=Phi(jk)+ALF*ZK(jk)
            RES(jk)=RES(jk)-ALF*VK(jk)
            RESN=RESN+ABS(RES(jk))
          END DO
        END DO


!... check CONVERGENCE
!
        RSM=RESN/(RESZERO+1.0D-20)
       IF( RSM .LT. RESMAX )RETURN

!        IF( RSM .LT. RESMAX )THEN
!      PRINT*,"NUMBER OF ITERATIONS=",N

!         RETURN
!         END IF
      END DO
!------------------------------------------------------------------------
! END ITERATION
!------------------------------------------------------------------------
      RETURN
      END SUBROUTINE 


      SUBROUTINE WRITER_BEACH()
      REAL :: Qh, Ug, Vg
      REAL, DIMENSION(JS:JE,KS:KE) ::  zwrite,uwrite,vwrite
      REAL, DIMENSION(JS:JE,KS:KE) :: uenf,venf,UENFQ,VENFQ
      INTEGER, DIMENSION(2):: IZM
      REAL :: FMA
      INTEGER,PARAMETER::MA=60      ! number of time step for averaging

    !  PRINT *,m,  SLO(3,1000)
      ZWRITE=-1.
      FMA=FLOAT(MA)
! ================================================================
! SEA LEVEL AVG. FOR EACH MIN. AT STATION, (T=2sec SO 30 TIME STEP)
! ================================================================
      IF(MOD(M, MA) .EQ. 0) THEN   
        DO I=1,IS
          ZSU(I)=SLO(JMS(I),KMS(I))+ZSU(I)
            ZSU(I)=ZSU(I)/FMA
            UNITS=100+i
            WRITE(UNITS,*)ZSU(I)
            ZSU(I)=0.0
        END DO
      ELSE
        DO I=1,IS
          ZSU(I)=SLO(JMS(I),KMS(I))+ZSU(I)
        END DO
      END IF
! ================================================================
!if(mod(m,MA).eq.0)then
!DO I=1,IS
!UNITS=100+i
!WRITE(UNITS,*)SLO(JMS(I),KMS(I))
!end do
!END IF
!***********************************
!WRITE EVERY MINUTE ALONG N-S 129W CROSSECTION
!FROM 39.5N TO 40.5N
!IF(mod(m,MA).eq.0)THEN
!	SUME=0.
!	DO K=646,796
!	SUME=SUME+SLO(3331,K)*UO(3331,K)*H(3331,K)
!	END DO
!	SUME=SUME/150.
!	WRITE(40,*)SUME
!END IF
!=============================================
!WRITE EVERY MINUTE ALONG W-E 39.5N CROSSECTION
!FROM 129W TO 124W
!IF(mod(m,MA).eq.0)THEN
!	SUME=0.
!	DO J=3331,3481
!	SUME=SUME+SLO(J,646)*VO(J,646)*H(J,646)
!	END DO
!	SUME=SUME/150.
!	WRITE(41,*)SUME
!END IF
!================================================
!WRITE EVERY MINUTE ALONG W-E 44.5N CROSSECTION
!FROM 129W TO 124W
!IF(mod(m,MA).eq.0)THEN
!	SUME=0.
!	DO J=3331,3481
!	SUME=SUME+SLO(J,796)*VO(J,796)*H(J,796)
!	END DO
!	SUME=SUME/150.
!	WRITE(42,*)SUME
!END IF
!        if(mod(m,4*MA).eq.0)then
           if(mod(m,10).eq.0)then
!...
 PRINT *,M, 'SLO(1000,800)=',SLO(1000,800)
           PRINT*,"MAXVALUE= ", MAXVAL(SLO,MASK=(CELL.eq.1)),&
         "TIME STEP= ",M
         PRINT*,"MAX_LOCATION= ", MAXLOC(SLO,MASK=(CELL.eq.1))
            IZM = MAXLOC(SLO,MASK=(CELL.eq.1))
       PRINT*,  "DEPTH_AT_MAX_LOCATION= ", H( IZM(1), IZM(2))

 !      DO j=JS,Je

 !        write (14,*) UO(J,1500)

 !      write (15,*) SLO(J,1500)

 !       END DO
        ! DO J=4,JE-3
        do k=ks+1,ke
        do j=js,je-1
          UG=0.5*(UO(J,K)+UO(J+1,K))*H(J,K)
        VG=0.5*(VO(J,K)+VO(J,K-1))*H(J,K)
        QH=Q(J,K)*0.5
        UENF(J,K)=SLO(J,K)*UG/100.  !ENERGY FLUX
         VENF(J,K)=SLO(J,K)*VG/100.
         UENFQ(J,K)=UG*QH         !flux of energy due to nonhydrostatic pressure
         VENFQ(J,K)=VG*QH
           zwrite(j,k)=slo(j,k)
  IF(CELL(j,k).EQ.0.)zwrite(j,k)=-1.
 ! IF(CELL(j,k).EQ.1.AND.ETO(J,K).GT.H(J,K))zwrite(j,k)=5.

        uwrite(J,K)=UO(J,K)
        VWRITE(J,K)=VO(J,K)
        end do
        end do
  !      	DO K=31,34
  !	ZWRITE(30,K)=10.
  !     END DO
  !             ZWRITE(31,32)=10.
            !   DO K=1,KE,4

      !    write (16,*) (zwrite(J,K),J=1,JE,4)
 !         write (16,*) (SLO(J,K),J=1,JE)
      !     write (17,*) (uwrite(J,K),J=1,JE,4)
      !    write (18,*) (vwrite(J,K),J=1,JE,4)
      !   write (19,*) (uenf(J,K),J=1,JE,4)
      !    write (20,*) (venf(J,K),J=1,JE,4)

         ! END DO
         ! END DO
          write (20) ((venf(J,K),J=1,JE,4),K=1,KE,4)
          write (19) ((uenf(J,K),J=1,JE,4),K=1,KE,4)
          write (18) ((vwrite(J,K),J=1,JE,4),K=1,KE,4)
           write (17) ((uwrite(J,K),J=1,JE,4),K=1,KE,4)
          write (16) ((zwrite(J,K),J=1,JE,4),K=1,KE,4)

           write (43) ((uenfq(J,K),J=1,JE,4),K=1,KE,4)
           write (44) ((venfq(J,K),J=1,JE,4),K=1,KE,4)


 END IF

      end subroutine

      subroutine WRITER_MAX

       ! open(unit=21,file='ampmax.dat',status='unknown')
      DO K=KS,KE 
        DO J=JS,JE
          IF (cell(j, k) .eq. 0)ZMAX(j, k)=-1
        END DO
      END DO

      write (21) ((ZMAX(J,K),J=1,JE,2),K=1,KE,2)


       end subroutine




  END  program SLOSH




